"""In this test suite, we load the allowlist.json and run through all the tests based
on what expected inputs and return types are provided by the json file.
"""

# stdlib
from itertools import product
import json
import os
from pathlib import Path
import platform
import random
import sys
import time
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List as ListType
from typing import Union

# third party
from packaging import version
import pytest
from pytest import approx
import torch as th

# syft absolute
import syft as sy
from syft.core.pointer.pointer import Pointer
from syft.lib.python import List
from syft.lib.python import String
from syft.lib.python.primitive_factory import PrimitiveFactory
from syft.lib.python.primitive_factory import isprimitive
from syft.lib.python.primitive_interface import PyPrimitive
from syft.lib.torch import allowlist
from syft.lib.torch.return_types import types_fields
from syft.lib.torch.tensor_util import TORCH_STR_DTYPE
from syft.lib.util import full_name_with_qualname

TORCH_VERSION = version.parse(th.__version__.split("+")[0])
py_ver = sys.version_info
PYTHON_VERSION = version.parse(f"{py_ver.major}.{py_ver.minor}")
OS_NAME = platform.system().lower()


# Allow List Test Generation
#
# The purpose of this code is to gather all of the explicitly defined classes,
# attributes, methods and properties that are can be executed remotely via a pointer.
# The vast majority of types and inputs are tensors or collections of numeric values
# so for many cases simply enumerating all possible supported numeric data types
# and using them as both Tensors and inputs covers a lot of the functionality.
# However there are instances where specific methods or properties either do not
# accept input, have specific input requirements or kwargs, or give a variety of output
# types. Additionally some methods exist or behave differently depending on the version
# of the library that we are testing. To support all of this added complexity we start
# with the simplest version which is the "attribute path" to "return type" mapping aka:
#
# allowlist["torch.Tensor.__add__"] = "torch.Tensor"
#
# We then check to see if a explicit configuration is set for the attribute path in
# allowlist_test.json and if so we use the configuration from there.
# To keep the custom definitions DRY we use a concept called "profiles" which are just
# a key lookup in the main JSON tree. Many configuration options also support the
# concept of a key lookup for example you will see lots of "all" which is used to map
# to a whole pre-defined set of supported data types, or inputs. By utilising this
# approach we can prevent the need to maintain a large list of expected exceptions
# which increases the likelihood tests are skipped by accident and makes it very hard
# to track exactly which data types and inputs are supported for a given attribute.
#
# Once we understand what can be tested we construct a product or combination of all
# possible options using the following required variables for each test run:
#
# tensor_type, op_name, self_tensor, _args, is_property, return_type
#
# tensor_type:  this is the data type we convert our test tensor to e.g. uint8
# op_name:      this is the method or property we are testing e.g. __add__
# self_tensor:  this is the tensor we are operating when doing the test
# _args:        this is the input used on the method where "self" represents the
#               self_tensor and None represents no input
# is_property:  this helps understand if the attribute can be called or not
# return_type:  this helps verify that the return type matches the expected return type

# this handles instances where the allow list provides more meta information
def get_return_type(support_dict: Union[str, Dict[str, str]]) -> str:
    if isinstance(support_dict, str):
        return support_dict
    else:
        return support_dict["return_type"]


# this allows us to skip tests based on a few loose matching rules for example:
#
# "skip": [{ "data_types": ["bool"], "inputs": [0, 1] }]
#
# This rule set will match on any test which is for the data type bool and has the
# inputs 0 or 1. Since the primary variation of test combinations relates to their
# tensors, inputs and the data types associated with them
def check_skip(
    combination: ListType,
    skip_rule: Dict[str, Any],
    lib_version: Union[version.Version, version.LegacyVersion],
) -> bool:
    combination_dict = {
        "data_types": combination[0],
        "tensors": combination[2],
        "inputs": combination[3],
    }

    # we will assume the skip rule applies and try to prove it doesn't by finding a
    # condition which we cannot satisfy
    applies = True
    for key in skip_rule.keys():
        # handle the version checks later
        if key.endswith("_version"):
            continue
        if key in combination_dict:
            # first check it is not equal since then the rule would apply
            if combination_dict[key] != skip_rule[key]:
                # second check if the rule is a list of rules in which case we
                # need to check for the presence of our current combination value
                if isinstance(skip_rule[key], Iterable):
                    # we cant use x not in y because bool matches for 0 and 1 in lists
                    presence = any(
                        combination_dict[key] == item
                        and type(combination_dict[key]) == type(item)
                        for item in skip_rule[key]
                    )

                    # if we cant find a matching item of value and type then we dont
                    # match and we wont skip this test, if its True we match so far and
                    # can keep checking more rules
                    applies = presence
                    if not applies:
                        return applies

                else:
                    # we dont match and the rule is not a list of rules so we wont skip
                    # and can break
                    applies = False
                    return applies

    # skip rule versioning example:
    # "skip": [{ "data_types": ["float16"], "lte_version": "1.5.1" }]
    # this says skip float16 if the library is less than or equal to 1.5.1
    # say we are running 1.5.0 that means
    # torch==1.5.0 <= 1.5.1 so this rule applies and we could possibly need to skip

    # if there is a lte_version for the skip rule and our lib is NOT lte, less than or
    # equal to the version then we cant skip because the rule only prohibits lte
    if "lte_version" in skip_rule and (
        not lib_version <= version.parse(skip_rule["lte_version"])
    ):
        applies = False

    # if there is a gte_version for the skip rule and our lib is NOT gte, greater than
    # or equal to the version then  we cant skip because the rule only prohibits gte
    if "gte_version" in skip_rule and (
        not lib_version >= version.parse(skip_rule["gte_version"])
    ):
        applies = False

    # if we were unable to fail the skip rules we should have applies = True and will
    # skip this particular test combination
    return applies


BASIC_OPS = list()
BASIC_OPS_RETURN_TYPE = {}
# here we are loading up the true allowlist which means that we are only testing what
# can be used by the end user
for method, return_type_name_or_dict in allowlist.items():
    if method.startswith("torch.Tensor."):
        return_type = get_return_type(support_dict=return_type_name_or_dict)
        if return_type == "unknown":
            return_type = "torch.Tensor"
        method_name = method.split(".")[-1]
        BASIC_OPS.append(method_name)
        BASIC_OPS_RETURN_TYPE[method_name] = return_type

# load our custom configurations
try:
    with open(__file__.replace(".py", ".json"), "r") as f:
        TEST_JSON = json.loads(f.read())
except Exception as e:
    print(f"Exception {e} triggered")
    raise e

# we need a file to keep all the errors in that makes it easy to debug failures
TARGET_PLATFORM = f"{PYTHON_VERSION}_{TORCH_VERSION}_{OS_NAME}"
ERROR_FILE_PATH = os.path.abspath(
    Path(__file__) / "../../../.." / f"allowlist_test_errors_{TARGET_PLATFORM}.jsonl"
)

# we need a file to keep track of all the methods that are skipped or succeed
SUPPORT_FILE_PATH = os.path.abspath(
    Path(__file__) / "../../../.." / f"allowlist_test_support_{TARGET_PLATFORM}.jsonl"
)


# clear the file before running the tests
if os.path.exists(ERROR_FILE_PATH):
    # this one we can delete since we dont start writing until we are into the tests
    try:
        os.unlink(ERROR_FILE_PATH)
    except Exception as e:
        print(f"Exception {e} triggered")


# we are running many works in parallel and theres a race condition with deleting this
# file and then writing to it during the collection phase so we are going to just
# spread out the workers and only delete if the file isn't brand new
time.sleep(random.random() * 2)

if os.path.exists(SUPPORT_FILE_PATH):
    # we need to write during gathering so we need to delete this carefully
    try:
        file_stat = os.stat(SUPPORT_FILE_PATH)
        diff = time.time() - file_stat.st_mtime
        if diff > 0.1:
            # only delete on the first run
            for retry in range(5):
                try:
                    os.unlink(SUPPORT_FILE_PATH)
                    break
                except BaseException:
                    time.sleep(1)
    except Exception:
        print(f"Failed while trying to os.stat file {SUPPORT_FILE_PATH}")


# write test debug info to make it easy to debug long running tests with large output
def write_error_debug(debug_data: Dict[str, Any]) -> None:
    # save a file in the root project dir
    try:
        with open(ERROR_FILE_PATH, "a+") as f:
            f.write(f"{json.dumps(debug_data, default=str)}\n")
    except Exception as e:
        print(f"Exception {e} triggered")


# write the result of support for the test for creating a comprehensive report
def write_support_result(test_details: Dict[str, Any]) -> None:
    # save a file in the root project dir
    try:
        with open(SUPPORT_FILE_PATH, "a+") as f:
            f.write(f"{json.dumps(test_details, default=str)}\n")
    except Exception as e:
        print(f"Exception {e} triggered")


TEST_DATA = []
# we iterate over the allowlist and only override where explicitly defined
for op in BASIC_OPS:
    skip = []
    not_available = []
    deterministic = True
    if op not in TEST_JSON["tests"]["torch.Tensor"]:
        # there is no custom configuration so we will test all supported combinations
        dtypes = ["common"]
        test_tensors = ["tensor1", "tensor2"]
        test_inputs = ["all"]
        is_property = False
        return_type = BASIC_OPS_RETURN_TYPE[op]
        # continue  # dont try testing without testing rules
    else:
        meta = {}
        # if there is a profile use the settings from that first
        if "profile" in TEST_JSON["tests"]["torch.Tensor"][op]:
            profile = TEST_JSON["tests"]["torch.Tensor"][op]["profile"]
            meta.update(TEST_JSON["profiles"][profile])

        # now update with what ever extra settings are overridden on the method
        meta.update(TEST_JSON["tests"]["torch.Tensor"][op])
        dtypes = meta["data_types"]
        test_tensors = meta["tensors"]
        test_inputs = meta["inputs"]
        is_property = meta["property"]
        return_type = meta["return_type"]

        # grab skip rules
        if "skip" in meta:
            skip += meta["skip"]

        # grab not_available rules
        if "not_available" in meta:
            not_available += meta["not_available"]

        if "deterministic" in meta:
            deterministic = meta["deterministic"]

        # this is the minimum version of the library required to run this test
        # which should match the values in the actual allowlist.py
        if "min_version" in meta and TORCH_VERSION < version.parse(meta["min_version"]):
            # skip attributes which are not supported in the TORCH_VERSION
            continue

    # gather all the data types we want to test
    data_types = []
    for dtype in dtypes:
        if issubclass(type(dtype), str) and dtype in TEST_JSON["data_types"]:
            data_types += TEST_JSON["data_types"][dtype]
        else:
            data_types.append(dtype)

    # collect the tensors we want to test
    tensors = []
    for tensor in test_tensors:
        if issubclass(type(tensor), str) and tensor in TEST_JSON["tensors"]:
            tensors += TEST_JSON["tensors"][tensor]
        else:
            tensors.append(tensor)

    # get the list of inputs
    inputs = []
    for input in test_inputs:
        if issubclass(type(input), str) and input in TEST_JSON["inputs"]:
            inputs += TEST_JSON["inputs"][input]
        elif issubclass(type(input), dict):
            resolved_args = {}
            for k, v in input.items():  # type: ignore
                if issubclass(type(v), str) and v in TEST_JSON["inputs"]:
                    resolved_args[k] = TEST_JSON["inputs"][v]
                else:
                    resolved_args[k] = v
            inputs.append(resolved_args)
        else:
            inputs.append(input)

    combinations = list(
        product(
            data_types,
            [op],
            tensors,
            inputs,
            [is_property],
            [return_type],
            [deterministic],
        )
    )

    skipped_combinations = set()
    not_available_combinations = set()
    for combination in combinations:
        # skips are temporary
        for skip_rule in skip:
            if check_skip(
                combination=list(combination),
                skip_rule=skip_rule,
                lib_version=TORCH_VERSION,
            ):
                # we use str(combination) so that we can hash the entire combination
                skipped_combinations.add(str(combination))

        # not available are features we cant or wont test because they arent supported
        for na_rule in not_available:
            if check_skip(
                combination=list(combination),
                skip_rule=na_rule,
                lib_version=TORCH_VERSION,
            ):
                # we use str(combination) so that we can hash the entire combination
                not_available_combinations.add(str(combination))

    for combination in combinations:
        # we need to record the support for this combination, the key will be unique
        # and easy to match multiple entries of the same combination
        support_data = {}
        support_data["tensor_type"] = combination[0]
        support_data["op_name"] = combination[1]

        # we use str so that we can hash the entire combination with nested lists
        if (
            str(combination) not in skipped_combinations
            and str(combination) not in not_available_combinations
        ):
            TEST_DATA.append(combination)
        elif str(combination) in not_available_combinations:
            support_data["status"] = "not_available"
            write_support_result(support_data)
        elif str(combination) in skipped_combinations:
            support_data["status"] = "skip"
            write_support_result(support_data)


# if the environment variables below are set bigger than 1 we will split the TEST_DATA
# into parts so that these can be parallelized by different test runners or containers
# TEST_CHUNK = int(os.getenv("TEST_CHUNK", 1))
# TEST_CHUNKS = int(os.getenv("TEST_CHUNKS", 1))

# # chunk the tests
# if TEST_CHUNKS > 1:
#     chunk_size = math.ceil(len(TEST_DATA) / TEST_CHUNKS)
#     start_offset = (TEST_CHUNK - 1) * chunk_size
#     end_offset = start_offset + chunk_size
#     TEST_DATA = TEST_DATA[start_offset:end_offset]


@pytest.mark.torch
@pytest.mark.parametrize(
    "tensor_type, op_name, self_tensor, _args, is_property, return_type, deterministic",
    TEST_DATA,
)
def test_all_allowlisted_tensor_methods(
    tensor_type: str,
    op_name: str,
    self_tensor: ListType,
    _args: Union[str, ListType, Dict, bool, None],
    is_property: bool,
    return_type: str,
    deterministic: bool,
    client: sy.VirtualMachineClient,
) -> None:

    support_data = {}
    support_data["tensor_type"] = tensor_type
    support_data["op_name"] = op_name

    # this is used in bulk error reporting
    debug_data: Dict[str, Any] = {}
    debug_data["input"] = {
        "tensor_type": tensor_type,
        "op_name": op_name,
        "self_tensor": self_tensor,
        "_args": _args,
        "is_property": is_property,
        "return_type": return_type,
        "deterministic": deterministic,
    }

    try:
        # Step 2: Decide which type we're testing
        t_type = TORCH_STR_DTYPE[tensor_type]

        # Step 3: Create the object we're going to call a method on
        # NOTE: we need a second copy because some methods mutate tensor before we send
        requires_grad = False
        if op_name in ["backward", "retain_grad", "grad"]:
            requires_grad = True
        self_tensor, self_tensor_copy = (
            th.tensor(self_tensor, dtype=t_type, requires_grad=requires_grad),
            th.tensor(self_tensor, dtype=t_type, requires_grad=requires_grad),
        )

        # we dont have .id's by default anymore
        # self_tensor_copy.id = self_tensor.id  # type: ignore

        args: Union[ListType[Any], Dict[str, Any]]

        # Step 4: Create the arguments we're going to pass the method on
        if _args is None:
            args = []
        elif _args == "self":
            args = [th.tensor(self_tensor, dtype=t_type, requires_grad=requires_grad)]
        elif isinstance(_args, list):
            args = [th.tensor(_args, dtype=t_type)]
        elif isinstance(_args, dict):
            args = {}
            tuple_args = False
            if len(_args) > 0 and "ALL_TUPLE_ARGS_" in list(_args.keys())[0]:
                tuple_args = True
            for k, v in _args.items():
                arg_type = t_type
                real_k = k
                if isinstance(v, list):
                    if "_DTYPE_" in real_k:
                        parts = real_k.split("_DTYPE_")
                        real_k = parts[0]
                        v_dtype_attr = parts[1]
                        v_dtype = getattr(th, v_dtype_attr, None)
                        if v_dtype is not None:
                            arg_type = v_dtype

                    if real_k.startswith("LIST_"):
                        # use a normal list
                        real_k = real_k.replace("LIST_", "")
                        if real_k.startswith("TENSOR_"):
                            real_k = real_k.replace("TENSOR_", "")
                            args[real_k] = [th.tensor(t, dtype=arg_type) for t in v]
                        else:
                            args[real_k] = v
                    elif real_k.startswith("0d_"):
                        # make a 0d tensor
                        real_k = real_k.replace("0d_", "")
                        args[real_k] = th.tensor(v[0], dtype=arg_type)
                    else:
                        args[real_k] = th.tensor(v, dtype=arg_type)
                elif v == "self":
                    args[real_k] = [
                        th.tensor(
                            self_tensor, dtype=arg_type, requires_grad=requires_grad
                        )
                    ]
                else:
                    args[real_k] = v
            if tuple_args:
                args = list(args.values())
        else:
            args = [PrimitiveFactory.generate_primitive(value=_args, recurse=True)]

        try:
            _ = len(args)
        except Exception as e:
            err = (
                f"Args must be iterable so it can be spread with * into the method. {e}"
            )
            raise Exception(err)

        # Step 4: Get the method we're going to call
        # if op_name=="grad", we need to do more operations first
        if op_name == "grad":
            self_tensor.sum().backward()  # type: ignore
        target_op_method = getattr(self_tensor, op_name)

        # Step 5: Test to see whether this method and arguments combination is valid
        # in normal PyTorch. If it this is an invalid combination, abort the test
        try:
            if not is_property:
                upcasted_args: Union[dict, list]
                # this prevents things like syft.lib.python.bool.Bool from getting
                # treated as an Int locally but then failing on the upcast to
                # builtins.bool on the remote side
                if isinstance(args, dict):
                    upcasted_args = {}
                    for k, arg in args.items():
                        upcast_attr = getattr(arg, "upcast", None)
                        if upcast_attr is not None:
                            upcasted_args[k] = upcast_attr()
                        else:
                            upcasted_args[k] = arg
                else:
                    upcasted_args = []
                    for arg in args:
                        upcast_attr = getattr(arg, "upcast", None)
                        if upcast_attr is not None:
                            upcasted_args.append(upcast_attr())
                        else:
                            upcasted_args.append(arg)

                if issubclass(type(upcasted_args), dict):
                    target_result = target_op_method(**upcasted_args)
                else:
                    target_result = target_op_method(*upcasted_args)
            else:
                # we have a property and already have its value
                target_result = target_op_method

            debug_data["target_result"] = target_result
            debug_data["target_result_type"] = type(target_result)

        except Exception as e:
            error = (
                "Exception in allowlist suite. If this is an expected exception, update"
            )
            error += " the .json file to prevent this test combination."
            print(error)
            raise e

        # Step 6: Send our target tensor to alice.
        # NOTE: send the copy we haven't mutated
        xp = self_tensor_copy.send(client)

        # if op_name=="grad", we need to do more operations first
        if op_name == "grad":
            xp.sum().backward()

        argsp: ListType[Any] = []
        if len(args) > 0 and not is_property:
            if isinstance(args, dict):
                argsp = [
                    arg.send(client) if hasattr(arg, "send") else arg
                    for arg in args.values()
                ]
            else:
                argsp = [
                    arg.send(client) if hasattr(arg, "send") else arg for arg in args
                ]

        # Step 7: get the method on the pointer to alice we want to test
        op_method = getattr(xp, op_name, None)

        # Step 8: make sure the method exists
        assert op_method is not None

        # Step 9: Execute the method remotely
        if is_property:
            # we already have the result
            result = op_method
        else:
            result = op_method(*argsp)

        # Step 10: Ensure the method returned a pointer
        assert isinstance(result, Pointer)

        # Step 11: Get the result
        result_pointer_type = type(result)
        local_result = result.get()

        debug_data["local_result"] = local_result
        debug_data["local_result_type"] = type(local_result)

        # Step 12: If there are NaN values, set them to 0 (normal for division by 0)
        try:
            target_fqn = full_name_with_qualname(klass=type(target_result))
            if target_fqn.startswith("torch.return_types."):
                fields = types_fields[type(local_result)]
                for field in fields:
                    assert compare_tensors(
                        left=getattr(local_result, field, None),
                        right=getattr(target_result, field, None),
                    )
            elif (
                issubclass(type(local_result), (str, String))
                or not hasattr(target_result, "__len__")
                and (
                    isprimitive(value=target_result)
                    or issubclass(type(local_result), PyPrimitive)
                )
            ):
                # only do single value comparisons, do lists, tuples etc below in the else
                # check that it matches functionally
                if deterministic:
                    if issubclass(type(target_result), float):
                        assert local_result + target_result == approx(2 * target_result)
                    else:
                        assert local_result == target_result

                # convert target_result for type comparison below
                target_result = PrimitiveFactory.generate_primitive(value=target_result)
            else:
                if issubclass(type(target_result), tuple) or issubclass(
                    type(local_result), tuple
                ):
                    target_result = list(target_result)
                    local_result = list(local_result)

                # to handle tuple return types for now we just make sure everything
                # is in a list of at least 1 size and then iterate over it
                delist = False
                if type(local_result) is not list:
                    delist = True
                    local_result = [local_result]
                    target_result = [target_result]

                if deterministic:
                    for i, local_item in enumerate(local_result):
                        target_item = target_result[i]
                        if issubclass(type(target_item), th.Tensor) and issubclass(
                            type(local_item), th.Tensor
                        ):
                            assert compare_tensors(left=target_item, right=local_item)
                        else:
                            if not hasattr(local_item, "__len__") and not hasattr(
                                target_item, "__len__"
                            ):
                                assert local_item + target_item == approx(
                                    2 * local_item
                                )
                            else:
                                for left, right in zip(local_item, target_item):
                                    assert left + right == approx(2 * left)

                if delist:
                    # debox the tensors if they were not lists originally
                    local_result = local_result[0]
                    target_result = target_result[0]

            # make sure the return types match
            if isprimitive(value=target_result):
                target_result = PrimitiveFactory.generate_primitive(value=target_result)

            if isprimitive(value=local_result):
                local_result = PrimitiveFactory.generate_primitive(value=local_result)
            assert type(local_result) == type(target_result)

            # TODO: Fix this workaround for types that sometimes return Tensor tuples
            # we are getting back more than 1 return type so we need to fake it until
            # we add Union to the return types
            if (
                hasattr(local_result, "__len__")
                and not issubclass(type(local_result), (th.Tensor, str, String))
                and not target_fqn.startswith("torch.return_types.")
            ):
                if issubclass(type(local_result), (list, List)) and issubclass(
                    type(target_result), (list, List)
                ):
                    assert len(local_result) == len(target_result)
                    for left, right in zip(local_result, target_result):
                        assert type(left) == type(right)
                        if deterministic:
                            if issubclass(type(left), th.Tensor) and issubclass(
                                type(right), th.Tensor
                            ):
                                assert compare_tensors(left=left, right=right)
                            elif not hasattr(local_item, "__len__") and not hasattr(
                                target_item, "__len__"
                            ):
                                assert local_item + target_item == approx(
                                    2 * local_item
                                )
                            else:
                                for left, right in zip(local_item, target_item):
                                    assert left + right == approx(2 * left)
                else:
                    # TODO: Fix this when we find one
                    raise Exception("Unsupported Union return type")

            # make sure the return type matches the specified allowlist return type
            if not issubclass(type(local_result), th.Tensor):
                if target_fqn.startswith("torch.return_types."):
                    local_type = full_name_with_qualname(klass=type(local_result))
                else:
                    local_type = full_name_with_qualname(
                        klass=type(
                            PrimitiveFactory.generate_primitive(value=local_result)
                        )
                    )
                full_result_pointer_type = full_name_with_qualname(
                    klass=result_pointer_type
                )
                python_types = "syft.lib.python"
                if local_type.startswith(python_types) and return_type.startswith(
                    python_types
                ):
                    # python types seem to resolve as both int.Int and .Int causing issues
                    # in the match
                    assert local_type.split(".")[-1] == return_type.split(".")[-1]
                elif full_result_pointer_type.endswith("UnionPointer"):
                    union_part = local_type.rsplit(".", 1)[-1]
                    # check the returned value is part of the original expected Union
                    # Bool in syft.proxy.syft.lib.misc.union.BoolFloatIntUnionPointer
                    assert union_part in full_result_pointer_type
                    # check the result pointer type matches the expected test Union
                    assert full_result_pointer_type == return_type
                else:
                    assert local_type == return_type

            # Test Passes
            support_data["status"] = "pass"
            write_support_result(support_data)

        except Exception as e:
            error = "Exception in allowlist suite during final comparison."
            print(error)
            raise e

    except Exception as e:
        print(f"Test Exception: {e}")
        support_data["status"] = "fail"
        write_support_result(support_data)
        debug_data["exception"] = str(e)
        debug_data["exception_type"] = type(e)
        write_error_debug(debug_data)
        raise e


def compare_tensors(left: th.Tensor, right: th.Tensor) -> bool:
    try:
        # if they don't match we can try to remove NaN's
        if not (left == right).all():
            # Set all NaN to 0
            # If we have two tensors like
            # local = [Nan, 0, 1] and remote = [0, Nan, 1]
            # those are not equal
            # Tensor.isnan was added in torch 1.6
            # so we need to do torch.isnan(tensor)

            # th.isnan fails on some versions of torch or methods for
            # example the op T / t() and the data_type float16 on
            # RuntimeError: "ne_cpu" not implemented for 'Half'
            nan_mask = th.isnan(left)

            # check if any of the elements are actually NaN because
            # assigning to some masked positions fails on some versions
            # of torch and some methods so its best to avoid unless
            # actually needed
            if any(nan_mask.view(-1)):
                # Use the same mask for local and target
                left[nan_mask] = 0
                right[nan_mask] = 0

            # Step 13: Ensure we got the same result locally (using
            # normal pytorch) as we did remotely using Syft pointers to
            # communicate with remote torch objects
            assert (left == right).all()
        return True
    except Exception as e:
        # if the issue is with equality of the data_type we can try
        # comparing them with numpy
        if "eq_cpu" in str(e):
            assert (left.numpy() == right.numpy()).all()
            return True
        else:
            # otherwise lets just raise the exception
            raise e



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/lib/sympc/session.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.node.common import (
    client_pb2 as proto_dot_core_dot_node_dot_common_dot_client__pb2,
)
from syft.proto.lib.python import dict_pb2 as proto_dot_lib_dot_python_dot_dict__pb2

DESCRIPTOR = _descriptor.FileDescriptor(
    name="proto/lib/sympc/session.proto",
    package="syft.lib.sympc",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n\x1dproto/lib/sympc/session.proto\x12\x0esyft.lib.sympc\x1a#proto/core/node/common/client.proto\x1a\x1bproto/lib/python/dict.proto"\xe4\x01\n\nMPCSession\x12\x10\n\x08protocol\x18\x08 \x01(\x0c\x12\x0c\n\x04uuid\x18\x01 \x01(\x0c\x12\x0c\n\x04rank\x18\x02 \x01(\x05\x12\x11\n\tring_size\x18\x03 \x01(\x0c\x12\x12\n\nnr_parties\x18\x04 \x01(\x0c\x12%\n\x06\x63onfig\x18\x05 \x01(\x0b\x32\x15.syft.lib.python.Dict\x12.\n\x07parties\x18\x06 \x03(\x0b\x32\x1d.syft.core.node.common.Client\x12*\n\x03ttp\x18\x07 \x01(\x0b\x32\x1d.syft.core.node.common.Clientb\x06proto3',
    dependencies=[
        proto_dot_core_dot_node_dot_common_dot_client__pb2.DESCRIPTOR,
        proto_dot_lib_dot_python_dot_dict__pb2.DESCRIPTOR,
    ],
)


_MPCSESSION = _descriptor.Descriptor(
    name="MPCSession",
    full_name="syft.lib.sympc.MPCSession",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="protocol",
            full_name="syft.lib.sympc.MPCSession.protocol",
            index=0,
            number=8,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="uuid",
            full_name="syft.lib.sympc.MPCSession.uuid",
            index=1,
            number=1,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="rank",
            full_name="syft.lib.sympc.MPCSession.rank",
            index=2,
            number=2,
            type=5,
            cpp_type=1,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="ring_size",
            full_name="syft.lib.sympc.MPCSession.ring_size",
            index=3,
            number=3,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="nr_parties",
            full_name="syft.lib.sympc.MPCSession.nr_parties",
            index=4,
            number=4,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="config",
            full_name="syft.lib.sympc.MPCSession.config",
            index=5,
            number=5,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="parties",
            full_name="syft.lib.sympc.MPCSession.parties",
            index=6,
            number=6,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="ttp",
            full_name="syft.lib.sympc.MPCSession.ttp",
            index=7,
            number=7,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=116,
    serialized_end=344,
)

_MPCSESSION.fields_by_name[
    "config"
].message_type = proto_dot_lib_dot_python_dot_dict__pb2._DICT
_MPCSESSION.fields_by_name[
    "parties"
].message_type = proto_dot_core_dot_node_dot_common_dot_client__pb2._CLIENT
_MPCSESSION.fields_by_name[
    "ttp"
].message_type = proto_dot_core_dot_node_dot_common_dot_client__pb2._CLIENT
DESCRIPTOR.message_types_by_name["MPCSession"] = _MPCSESSION
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

MPCSession = _reflection.GeneratedProtocolMessageType(
    "MPCSession",
    (_message.Message,),
    {
        "DESCRIPTOR": _MPCSESSION,
        "__module__": "proto.lib.sympc.session_pb2"
        # @@protoc_insertion_point(class_scope:syft.lib.sympc.MPCSession)
    },
)
_sym_db.RegisterMessage(MPCSession)


# @@protoc_insertion_point(module_scope)



# stdlib
import time
from typing import Any
from typing import Dict as DictType
from typing import List
from typing import Optional

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from nacl.signing import VerifyKey

# syft relative
from ..... import deserialize
from ..... import serialize
from .....lib.python import Dict
from .....lib.python.util import downcast
from .....lib.python.util import upcast
from .....logger import debug
from .....logger import error
from .....proto.core.node.domain.service.request_handler_message_pb2 import (
    GetAllRequestHandlersMessage as GetAllRequestHandlersMessage_PB,
)
from .....proto.core.node.domain.service.request_handler_message_pb2 import (
    GetAllRequestHandlersResponseMessage as GetAllRequestHandlersResponseMessage_PB,
)
from .....proto.core.node.domain.service.request_handler_message_pb2 import (
    UpdateRequestHandlerMessage as UpdateRequestHandlerMessage_PB,
)
from .....util import traceback_and_raise
from ....common import UID
from ....common.message import ImmediateSyftMessageWithReply
from ....common.message import ImmediateSyftMessageWithoutReply
from ....common.serde.serializable import bind_protobuf
from ....io.address import Address
from ...abstract.node import AbstractNode
from ...common.service.node_service import ImmediateNodeServiceWithoutReply


@bind_protobuf
class UpdateRequestHandlerMessage(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        handler: DictType[str, Any],
        address: Address,
        msg_id: Optional[UID] = None,
        keep: bool = True,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.handler = handler
        self.keep = keep

    def _object2proto(self) -> UpdateRequestHandlerMessage_PB:
        """Returns a protobuf serialization of self.

        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.

        :return: returns a protobuf object
        :rtype: UpdateRequestHandlerMessage_PB

        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """

        return UpdateRequestHandlerMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            handler=downcast(value=self.handler)._object2proto(),
            keep=self.keep,
        )

    @staticmethod
    def _proto2object(
        proto: UpdateRequestHandlerMessage_PB,
    ) -> "UpdateRequestHandlerMessage":
        """Creates a UpdateRequestHandlerMessage from a protobuf

        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.

        :return: returns an instance of UpdateRequestHandlerMessage
        :rtype: UpdateRequestHandlerMessage

        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return UpdateRequestHandlerMessage(
            msg_id=deserialize(blob=proto.msg_id),
            address=deserialize(blob=proto.address),
            handler=upcast(value=Dict._proto2object(proto=proto.handler)),
            keep=proto.keep,
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType

        """

        return UpdateRequestHandlerMessage_PB


@bind_protobuf
class GetAllRequestHandlersMessage(ImmediateSyftMessageWithReply):
    def __init__(
        self, address: Address, reply_to: Address, msg_id: Optional[UID] = None
    ):
        super().__init__(address=address, msg_id=msg_id, reply_to=reply_to)

    def _object2proto(self) -> GetAllRequestHandlersMessage_PB:
        """Returns a protobuf serialization of self.

        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.

        :return: returns a protobuf object
        :rtype: GetAllRequestHandlersMessage_PB

        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return GetAllRequestHandlersMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            reply_to=serialize(self.reply_to),
        )

    @staticmethod
    def _proto2object(
        proto: GetAllRequestHandlersMessage_PB,
    ) -> "GetAllRequestHandlersMessage":
        """Creates a GetAllRequestHandlersMessage from a protobuf

        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.

        :return: returns an instance of GetAllRequestHandlersMessage
        :rtype: GetAllRequestHandlersMessage

        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetAllRequestHandlersMessage(
            msg_id=deserialize(blob=proto.msg_id),
            address=deserialize(blob=proto.address),
            reply_to=deserialize(blob=proto.reply_to),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType

        """

        return GetAllRequestHandlersMessage_PB


@bind_protobuf
class GetAllRequestHandlersResponseMessage(ImmediateSyftMessageWithoutReply):
    def __init__(
        self,
        handlers: List[DictType],
        address: Address,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address=address, msg_id=msg_id)
        self.handlers = handlers

    def _object2proto(self) -> GetAllRequestHandlersResponseMessage_PB:
        """Returns a protobuf serialization of self.

        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.

        :return: returns a protobuf object
        :rtype: GetAllRequestHandlersResponseMessage_PB

        .. note::
            This method is purely an internal method. Please use serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        # For handler["created_time"], it's a large number. In order to keep it's precision
        # when serde, we turn it to string, and then turn it back to float in _proto2object.
        handlers = [h.copy() for h in self.handlers]
        for handler in handlers:
            if "created_time" in handler:
                handler["created_time"] = str(handler["created_time"])

        return GetAllRequestHandlersResponseMessage_PB(
            msg_id=serialize(self.id),
            address=serialize(self.address),
            handlers=list(map(lambda x: downcast(value=x)._object2proto(), handlers)),
        )

    @staticmethod
    def _proto2object(
        proto: GetAllRequestHandlersResponseMessage_PB,
    ) -> "GetAllRequestHandlersResponseMessage":
        """Creates a GetAllRequestHandlersResponseMessage from a protobuf

        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.

        :return: returns an instance of GetAllRequestHandlersResponseMessage
        :rtype: GetAllRequestHandlersResponseMessage

        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """
        handlers = [upcast(value=Dict._proto2object(proto=x)) for x in proto.handlers]
        for handler in handlers:
            if "created_time" in handler:
                handler["created_time"] = float(handler["created_time"])

        return GetAllRequestHandlersResponseMessage(
            msg_id=deserialize(blob=proto.msg_id),
            address=deserialize(blob=proto.address),
            handlers=handlers,
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type

        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.

        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.

        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType

        """

        return GetAllRequestHandlersResponseMessage_PB


class UpdateRequestHandlerService(ImmediateNodeServiceWithoutReply):
    @staticmethod
    def message_handler_types() -> List[type]:
        return [UpdateRequestHandlerMessage]

    @staticmethod
    def process(
        node: AbstractNode,
        msg: UpdateRequestHandlerMessage,
        verify_key: Optional[VerifyKey] = None,
    ) -> None:
        if verify_key is None:
            traceback_and_raise(
                ValueError(
                    "Can't process Request service without a given " "verification key"
                )
            )
        if verify_key == node.root_verify_key:
            replacement_handlers = []

            # find if there exists a handler match the handler passed in
            existing_handlers = getattr(node, "request_handlers", None)
            debug(f"> Updating Request Handlers with existing: {existing_handlers}")
            if existing_handlers is not None:
                matched = None
                for existing_handler in existing_handlers:
                    # we match two handlers according to their tags
                    if existing_handler["tags"] == msg.handler["tags"]:
                        matched = existing_handler
                        # if an existing_handler match the passed in handler,
                        # we ignore it in for loop
                        continue
                    else:
                        # if an existing_handler does not match the passed in
                        # handler, we keep it
                        replacement_handlers.append(existing_handler)

                if msg.keep:
                    msg.handler["created_time"] = time.time()
                    replacement_handlers.append(msg.handler)
                    if matched is not None:
                        debug(
                            f"> Replacing a Request Handler {matched} with: {msg.handler}"
                        )
                    else:
                        debug(f"> Adding a Request Handler {msg.handler}")
                else:
                    debug(f"> Removing a Request Handler with: {msg.handler}")

                setattr(node, "request_handlers", replacement_handlers)
                debug(f"> Finished Updating Request Handlers with: {existing_handlers}")
            else:
                error(f"> Node has no Request Handlers attribute: {type(node)}")

        return


class GetAllRequestHandlersService(ImmediateNodeServiceWithoutReply):
    @staticmethod
    def message_handler_types() -> List[type]:
        return [GetAllRequestHandlersMessage]

    @staticmethod
    def process(
        node: AbstractNode,
        msg: GetAllRequestHandlersMessage,
        verify_key: Optional[VerifyKey] = None,
    ) -> GetAllRequestHandlersResponseMessage:

        if verify_key is None:
            traceback_and_raise(
                ValueError(
                    "Can't process Request service without a given " "verification key"
                )
            )

        handlers: List[DictType[str, Any]] = []
        if verify_key == node.root_verify_key:
            existing_handlers = getattr(node, "request_handlers", None)
            debug(f"> Getting all Existing Request Handlers: {existing_handlers}")
            if existing_handlers is not None:
                handlers = existing_handlers

        return GetAllRequestHandlersResponseMessage(
            handlers=handlers, address=msg.reply_to
        )



# stdlib
from typing import Any
from typing import Callable as CallableT
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

# syft relative
from .. import ast
from .. import lib
from ..core.common.pointer import AbstractPointer
from ..core.node.common.action.get_or_set_static_attribute_action import (
    GetSetStaticAttributeAction,
)
from ..core.node.common.action.get_or_set_static_attribute_action import (
    StaticAttributeAction,
)
from ..logger import traceback_and_raise


class StaticAttribute(ast.attribute.Attribute):
    """A method, function, or constructor which can be directly executed"""

    def __init__(
        self,
        parent: ast.attribute.Attribute,
        path_and_name: str,
        return_type_name: Optional[str] = None,
        client: Optional[Any] = None,
    ):
        super().__init__(
            path_and_name=path_and_name,
            return_type_name=return_type_name,
            client=client,
            parent=parent,
        )

    def get_remote_value(self) -> AbstractPointer:
        if self.path_and_name is None:
            traceback_and_raise(
                ValueError("Can't execute remote get if path is not specified.")
            )

        if self.client is None:
            if self.client is None:
                traceback_and_raise(
                    ValueError(
                        "Can't get remote enum attribute if there is no client"
                        "set to get it from"
                    )
                )

        return_tensor_type_pointer_type = self.client.lib_ast.query(
            path=self.return_type_name
        ).pointer_type

        ptr = return_tensor_type_pointer_type(client=self.client)

        msg = GetSetStaticAttributeAction(
            path=self.path_and_name,
            id_at_location=ptr.id_at_location,
            address=self.client.address,
            action=StaticAttributeAction.GET,
        )
        self.client.send_immediate_msg_without_reply(msg=msg)
        return ptr

    def solve_get_value(self) -> Any:
        self.apply_node_changes()

        if self.path_and_name is None:
            raise ValueError("path_and_none should not be None")

        return getattr(self.parent.object_ref, self.path_and_name.rsplit(".")[-1])

    def solve_set_value(self, set_value: Any) -> None:
        self.apply_node_changes()

        if self.path_and_name is None:
            raise ValueError("path_and_none should not be None")

        setattr(self.parent.object_ref, self.path_and_name.rsplit(".")[-1], set_value)

    def set_remote_value(self, set_arg: Any) -> None:
        if self.client is None:
            raise ValueError(
                "MAKE PROPER SCHEMA - Can't get remote value if there is no remote "
                "client"
            )

        if self.path_and_name is None:
            raise ValueError("MAKE PROPER SCHEMA")

        resolved_pointer_type = self.client.lib_ast.query(self.return_type_name)
        result = resolved_pointer_type.pointer_type(client=self.client)
        result_id_at_location = getattr(result, "id_at_location", None)

        downcasted_set_arg = lib.python.util.downcast(set_arg)
        downcasted_set_arg_ptr = downcasted_set_arg.send(self.client)

        cmd = GetSetStaticAttributeAction(
            path=self.path_and_name,
            id_at_location=result_id_at_location,
            address=self.client.address,
            action=StaticAttributeAction.SET,
            set_arg=downcasted_set_arg_ptr,
        )
        self.client.send_immediate_msg_without_reply(msg=cmd)
        return result

    def __call__(  # type: ignore
        self, action: StaticAttributeAction
    ) -> Optional[Union["ast.callable.Callable", CallableT]]:
        raise ValueError("MAKE PROPER SCHEMA, THIS SHOULD NEVER BE CALLED")

    def add_path(self, *args: List[Any], **kwargs: Dict[Any, Any]) -> None:  # type: ignore
        raise ValueError("MAKE PROPER SCHEMA")



# stdlib
from typing import Any

# third party
import torch as th

# syft absolute
import syft as sy
from syft.lib.python.collections import OrderedDict


def test_linear_module() -> None:
    alice = sy.VirtualMachine()
    alice_client = alice.get_root_client()

    # Linear
    fc = th.nn.Linear(4, 2)

    # send
    fc_ptr = fc.send(alice_client)

    # remote call
    res_ptr = fc_ptr(th.rand([1, 4]))

    assert res_ptr.get().shape == th.Size((1, 2))

    # remote update state dict
    sd2 = OrderedDict(th.nn.Linear(4, 2).state_dict())
    sd2_ptr = sd2.send(alice_client)
    fc_ptr.load_state_dict(sd2_ptr)

    # get
    remote_sd2 = fc_ptr.get().state_dict()
    assert (remote_sd2["weight"] == sd2["weight"]).all()
    assert (remote_sd2["bias"] == sd2["bias"]).all()


def test_relu_module() -> None:
    alice = sy.VirtualMachine()
    alice_client = alice.get_root_client()

    # ReLU
    relu = th.nn.ReLU(inplace=True)

    # send
    relu_ptr = relu.send(alice_client)

    # remote call
    rand_data = th.rand([1, 4])
    res_ptr = relu_ptr(rand_data)
    rand_output = res_ptr.get()
    assert rand_output.shape == th.Size((1, 4))

    relu2 = relu_ptr.get()
    assert type(relu) == type(relu2)
    rand_output2 = relu2(rand_data)
    assert (rand_output2 == rand_output).all()


def test_user_module() -> None:
    alice = sy.VirtualMachine()
    alice_client = alice.get_root_client()

    # user defined model
    class M(th.nn.Module):
        def __init__(self) -> None:
            super(M, self).__init__()
            self.fc1 = th.nn.Linear(4, 2)
            self.fc2 = th.nn.Linear(2, 1)

        def forward(self, x: Any) -> Any:
            x = self.fc1(x)
            x = self.fc2(x)
            return x

    m = M()

    # send
    m_ptr = m.send(alice_client)

    # remote update state dict
    sd = OrderedDict(M().state_dict())
    sd_ptr = sd.send(alice_client)
    m_ptr.load_state_dict(sd_ptr)

    # get
    sd2 = m_ptr.get().state_dict()

    assert (sd["fc1.weight"] == sd2["fc1.weight"]).all()
    assert (sd["fc1.bias"] == sd2["fc1.bias"]).all()
    assert (sd["fc2.weight"] == sd2["fc2.weight"]).all()
    assert (sd["fc2.bias"] == sd2["fc2.bias"]).all()



# third party
import pytest

# syft absolute
from syft.lib.python.float import Float

test_float = Float(42.5)
python_float = 42.5
other = Float(42.5)


def test_id_abs():
    res = test_float.__abs__()
    py_res = python_float.__abs__()

    assert res == py_res
    assert res.id
    assert res.id != test_float.id


def test_id_add():
    res = test_float.__add__(other)
    py_res = python_float.__add__(other)

    assert res == py_res
    assert res.id
    assert res.id != test_float.id


def test_id_bool():
    res = test_float.__bool__()
    py_res = python_float.__bool__()

    assert res == py_res
    assert res.id
    assert res.id != test_float.id


def test_id_divmod():
    r1, q1 = test_float.__divmod__(other)
    r2, q2 = python_float.__divmod__(other)

    assert r1 == r2
    assert q1 == q2
    assert r1.id
    assert q1.id
    assert r1.id != test_float.id
    assert q1.id != test_float.id
    assert r1.id != q1.id


def test_id_eq():
    res = test_float.__eq__(other)
    py_res = python_float.__eq__(other)

    assert res == py_res
    assert res.id
    assert res.id != test_float.id


def test_id_floordiv():
    res = test_float.__floordiv__(other)
    py_res = python_float.__floordiv__(other)

    assert res == py_res
    assert res.id
    assert res.id != test_float.id


def test_id_ge():
    res = test_float.__ge__(other)
    py_res = python_float.__ge__(other)

    assert res == py_res
    assert res.id
    assert res.id != test_float.id


def test_id_gt():
    res = test_float.__gt__(other)
    py_res = python_float.__gt__(other)

    assert res == py_res
    assert res.id
    assert res.id != test_float.id


def test_id_iadd():
    res = test_float.__iadd__(other)
    assert res.id == test_float.id


def test_id_ifloordiv():
    res = test_float.__ifloordiv__(other)
    assert res.id == test_float.id


def test_id_imod():
    res = test_float.__imod__(other)
    assert res.id == test_float.id


def test_id_imul():
    res = test_float.__imul__(other)
    assert res.id == test_float.id


def test_id_ipow():
    res = test_float.__ipow__(other)
    assert res.id == test_float.id


def test_id_isub():
    res = test_float.__isub__(other)
    assert res.id == test_float.id


def test_id_itruediv():
    res = test_float.__itruediv__(other)
    assert res.id == test_float.id


def test_id_le():
    res = test_float.__le__(other)
    py_res = python_float.__le__(other)

    assert res == py_res
    assert res.id
    assert res.id != test_float.id


def test_id_lt():
    res = test_float.__lt__(other)
    py_res = python_float.__lt__(other)

    assert res == py_res
    assert res.id
    assert res.id != test_float.id


def test_id_mod():
    res = test_float.__mod__(other)
    py_res = python_float.__mod__(other)

    assert res == py_res
    assert res.id
    assert res.id != test_float.id


def test_id_mul():
    res = test_float.__mul__(other)
    py_res = python_float.__mul__(other)

    assert res == py_res
    assert res.id
    assert res.id != test_float.id


def test_id_ne():
    res = test_float.__ne__(other)
    py_res = python_float.__ne__(other)

    assert res == py_res
    assert res.id
    assert res.id != test_float.id


def test_id_neg():
    res = test_float.__neg__()
    py_res = python_float.__neg__()

    assert res == py_res
    assert res.id
    assert res.id != test_float.id


def test_id_pow():
    res = test_float.__pow__(other)
    py_res = python_float.__pow__(other)

    assert res == py_res
    assert res.id
    assert res.id != test_float.id


def test_id_radd():
    res = test_float.__radd__(other)
    py_res = python_float.__radd__(other)

    assert res == py_res
    assert res.id
    assert res.id != test_float.id


def test_id_rdivmod():
    r1, q1 = test_float.__rdivmod__(other)
    r2, q2 = python_float.__rdivmod__(other)

    assert r1 == r2
    assert q1 == q2

    assert r1.id != q1.id
    assert r1.id != test_float.id
    assert q1.id != test_float.id


def test_id_rfloordiv():
    res = test_float.__rfloordiv__(other)
    py_res = python_float.__rfloordiv__(other)

    assert res == py_res
    assert res.id
    assert res.id != test_float.id


def test_id_rmod():
    res = test_float.__rmod__(other)
    py_res = python_float.__rmod__(other)

    assert res == py_res
    assert res.id
    assert res.id != test_float.id


def test_id_rmul():
    res = test_float.__rmul__(other)
    py_res = python_float.__rmul__(other)

    assert res == py_res
    assert res.id
    assert res.id != test_float.id


def test_id_round():
    res = test_float.__round__()
    py_res = python_float.__round__()

    assert res == py_res
    assert res.id
    assert res.id != test_float.id


def test_id_rpow():
    res = test_float.__rpow__(other)
    py_res = python_float.__rpow__(other)

    assert res == py_res
    assert res.id
    assert res.id != test_float.id


def test_id_rsub():
    res = test_float.__rsub__(other)
    py_res = python_float.__rsub__(other)

    assert res == py_res
    assert res.id
    assert res.id != test_float.id


def test_id_rtruediv():
    res = test_float.__rtruediv__(other)
    py_res = python_float.__rtruediv__(other)

    assert res == py_res
    assert res.id
    assert res.id != test_float.id


def test_id_sub():
    res = test_float.__sub__(other)
    py_res = python_float.__sub__(other)

    assert res == py_res
    assert res.id
    assert res.id != test_float.id


def test_id_truediv():
    res = test_float.__truediv__(other)
    py_res = python_float.__truediv__(other)

    assert res == py_res
    assert res.id
    assert res.id != test_float.id


def test_id_trunc():
    res = test_float.__trunc__()
    py_res = python_float.__trunc__()

    assert res == py_res
    assert res.id
    assert res.id != test_float.id


def test_id_as_integer_ratio():
    u1, l1 = test_float.as_integer_ratio()
    u2, l2 = python_float.as_integer_ratio()

    assert u1 == u2
    assert l1 == l2
    assert u1.id != l1.id
    assert l1.id != test_float.id
    assert test_float.id != u1.id


@pytest.mark.xfail
def test_id_binary():
    # TODO finish this when we have bytes support
    assert True is False


def test_id_conjugate():
    res = test_float.conjugate()
    py_res = python_float.conjugate()

    assert res == py_res
    assert res.id
    assert res.id != test_float.id


def test_id_hex():
    res = test_float.hex()
    py_res = python_float.hex()

    assert res == py_res
    assert res.id
    assert res.id != test_float.id


def test_id_imag():
    res = test_float.imag
    py_res = python_float.imag

    assert res == py_res
    assert res.id
    assert res.id != test_float.id


def test_id_is_integer():
    res = test_float.is_integer()
    py_res = python_float.is_integer()

    assert res == py_res
    assert res.id
    assert res.id != test_float.id


def test_id_real():
    res = test_float.real
    py_res = python_float.real

    assert res == py_res
    assert res.id
    assert res.id != test_float.id


def test_upcast():
    assert Float(42.5).upcast() == 42.5
    assert type(Float(42.5).upcast()) is float



# stdlib
from typing import List as TypeList
from typing import Union

# syft relative
from ...lib.python.list import List
from .placeholder import PlaceHolder
from .state import State
from .state import StatePB


def wrap_model_params(parameters: Union[List, TypeList]) -> State:
    """
    Wraps list of tensors in State object.
    """
    state = State(
        state_placeholders=[
            PlaceHolder(id=n).instantiate(p) for n, p in enumerate(parameters)
        ]
    )
    return state


def deserialize_model_params(proto: bytes) -> List:
    """
    Deserializes binary string (State protobuf) to List of tensors.
    """
    state_pb = StatePB()
    state_pb.ParseFromString(proto)
    state: State = State._proto2object(state_pb)
    return List(state.tensors())



# stdlib
import logging
import os

# third party
import flask
from flask import Flask
from flask import Response

# syft absolute
from syft import serialize
from syft.core.common.message import SignedImmediateSyftMessageWithReply
from syft.core.common.message import SignedImmediateSyftMessageWithoutReply
from syft.core.common.serde.deserialize import _deserialize
from syft.core.node.network.network import Network
from syft.grid.services.signaling_service import PullSignalingService
from syft.grid.services.signaling_service import PushSignalingService
from syft.grid.services.signaling_service import RegisterDuetPeerService

log = logging.getLogger("werkzeug")
log.disabled = True


app = Flask(__name__)
network = Network(name="om-test-net")
network.immediate_services_without_reply.append(PushSignalingService)
network.immediate_services_with_reply.append(PullSignalingService)
network.immediate_services_with_reply.append(RegisterDuetPeerService)
network._register_services()  # re-register all services including SignalingService


@app.route("/metadata")
def metadata() -> flask.Response:
    metadata = network.get_metadata_for_client()
    metadata_proto = serialize(metadata)
    r = Response(
        response=metadata_proto.SerializeToString(),
        status=200,
    )
    r.headers["Content-Type"] = "application/octet-stream"
    return r


@app.route("/", methods=["POST"])
def post() -> flask.Response:
    data = flask.request.get_data()
    obj_msg = _deserialize(blob=data, from_bytes=True)
    if isinstance(obj_msg, SignedImmediateSyftMessageWithReply):
        reply = network.recv_immediate_msg_with_reply(msg=obj_msg)
        r = Response(response=serialize(reply, to_bytes=True), status=200)
        r.headers["Content-Type"] = "application/octet-stream"
        return r
    elif isinstance(obj_msg, SignedImmediateSyftMessageWithoutReply):
        network.recv_immediate_msg_without_reply(msg=obj_msg)
        r = Response(status=200)
        return r
    else:
        network.recv_eventual_msg_without_reply(msg=obj_msg)
        r = Response(status=200)
        return r


def run(port: int) -> None:
    global network
    PORT = os.getenv("PORT", port)
    app.debug = False
    app.use_reloader = False
    app.run(host="127.0.0.1", port=int(PORT))  # nosec



# third party
import pandas as pd

# syft relative
from ...generate_wrapper import GenerateWrapper
from ...lib.python.dict import Dict
from ...lib.python.primitive_factory import PrimitiveFactory
from ...proto.lib.pandas.frame_pb2 import PandasDataFrame as PandasDataFrame_PB


def object2proto(obj: pd.DataFrame) -> PandasDataFrame_PB:
    pd_dict = PrimitiveFactory.generate_primitive(value=obj.to_dict())
    dict_proto = pd_dict._object2proto()

    return PandasDataFrame_PB(
        id=dict_proto.id,
        dataframe=dict_proto,
    )


def proto2object(proto: PandasDataFrame_PB) -> pd.DataFrame:
    dataframe_dict = Dict._proto2object(proto=proto.dataframe)
    return pd.DataFrame.from_dict(dataframe_dict.upcast())


GenerateWrapper(
    wrapped_type=pd.DataFrame,
    import_path="pandas.DataFrame",
    protobuf_scheme=PandasDataFrame_PB,
    type_object2proto=object2proto,
    type_proto2object=proto2object,
)



# syft absolute
import syft as sy
from syft.lib.python.bool import Bool
from syft.proto.lib.python.bool_pb2 import Bool as Bool_PB


def test_serde() -> None:
    syft_bool = Bool(True)

    serialized = syft_bool._object2proto()

    assert isinstance(serialized, Bool_PB)

    deserialized = Bool._proto2object(proto=serialized)

    assert isinstance(deserialized, Bool)
    assert deserialized.id == syft_bool.id
    assert deserialized == syft_bool


def test_send(client: sy.VirtualMachineClient) -> None:
    syft_bool = Bool(5)
    ptr = syft_bool.send(client)
    # Check pointer type
    assert ptr.__class__.__name__ == "BoolPointer"

    # Check that we can get back the object
    res = ptr.get()
    assert res == syft_bool



# stdlib
import base64
from collections import OrderedDict
import json
import os
import time
from typing import Any
from typing import Dict as TypeDict
from typing import Generator
from typing import List as TypeList
from typing import Optional
from typing import Tuple as TypeTuple
from typing import Union as TypeUnion

# third party
import jwt
import numpy as np
import pytest
import requests
from syft_proto.execution.v1.plan_pb2 import Plan as PlanTorchscriptPB
import torch as th
from torchvision import datasets
from torchvision import transforms
from websocket import create_connection
from xprocess import ProcessStarter

# syft absolute
import syft as sy
from syft import deserialize
from syft import serialize
from syft.core.plan import Plan
from syft.core.plan.plan_builder import PLAN_BUILDER_VM
from syft.core.plan.plan_builder import ROOT_CLIENT
from syft.core.plan.plan_builder import make_plan
from syft.core.plan.translation.torchscript.plan import PlanTorchscript
from syft.core.plan.translation.torchscript.plan_translate import (
    translate as translate_to_ts,
)
from syft.federated import JSONDict
from syft.federated.fl_client import FLClient
from syft.federated.fl_job import FLJob
from syft.federated.model_centric_fl_client import ModelCentricFLClient
from syft.federated.model_serialization import deserialize_model_params
from syft.federated.model_serialization import wrap_model_params
from syft.lib.python.int import Int
from syft.lib.python.list import List
from syft.lib.torch.module import Module as SyModule
from syft.proto.core.plan.plan_pb2 import Plan as PlanPB
from syft.util import get_root_data_path

th.random.manual_seed(42)

here = os.path.dirname(__file__)
DOMAIN_PORT = 7000


@pytest.fixture
def pygrid_domain(xprocess: Any) -> Generator:
    class Starter(ProcessStarter):
        # startup pattern
        pattern = "Starting app"

        # command to start process
        pygrid_path = os.environ.get(
            "TEST_PYGRID_PATH", f"{here}/../../../../../pygrid"
        )
        domain_path = os.path.abspath(f"{pygrid_path}/apps/domain")
        database_file = f"{domain_path}/src/nodedatabase.db"
        if os.path.exists(database_file):
            os.unlink(database_file)
        args = [
            "python",
            f"{domain_path}/src/__main__.py",
            f"--port={DOMAIN_PORT}",
            "--start_local_db",
        ]

    # ensure process is running and return its logfile
    logfile = xprocess.ensure("pygrid_domain", Starter)

    yield logfile

    # clean up whole process tree afterwards
    xprocess.getinfo("pygrid_domain").terminate()


@pytest.mark.grid
def test_create_and_execute_plan_autograd(pygrid_domain: Any) -> None:
    fl_name = "mnist_autograd"
    plan, model = create_plan_autograd()

    plans = {"training_plan": plan}
    host_to_grid(plans, model, fl_name)

    plan_inputs = OrderedDict(
        {"xs": th.rand([64 * 3, 28 * 28]), "ys": th.randint(0, 10, [64 * 3, 10])}
    )
    plan_params_output_idx = [0, 1, 2, 3]
    model_param_type_size = sanity_check_hosted_plan(
        fl_name, model, plan_inputs, plan_params_output_idx
    )
    matches = [
        (th.Tensor, th.Size([100, 784])),
        (th.Tensor, th.Size([100])),
        (th.Tensor, th.Size([10, 100])),
        (th.Tensor, th.Size([10])),
    ]

    assert model_param_type_size == matches

    train_with_hosted_training_plan(fl_name, OrderedDict(), plan_params_output_idx)
    accuracy = check_resulting_model(fl_name, model)

    print(f"Model Centric Federated Learning Complete. Accuracy {accuracy:.2F}")
    assert accuracy > 0.05


@pytest.mark.grid
@pytest.mark.parametrize("plan_type", ["list", "torchscript"])
def test_create_and_execute_plan_mobile(pygrid_domain: Any, plan_type: str) -> None:
    fl_name = "mnist_mobile"
    plan, plan_ts, model = create_plan_mobile()

    plans = {
        "training_plan": plan,
        "training_plan:ts": plan_ts,
    }
    host_to_grid(plans, model, fl_name)

    bs = 64
    classes_num = 10
    plan_inputs = OrderedDict(
        {
            "xs": th.randn(bs, 28 * 28),
            "ys": th.nn.functional.one_hot(
                th.randint(0, classes_num, [bs]), classes_num
            ),
            "lr": th.tensor([0.1]),
            "batch_size": th.tensor([bs]),
        }
    )
    plan_output_params_idx = [2, 3, 4, 5]

    model_param_type_size = sanity_check_hosted_plan(
        fl_name, model, plan_inputs, plan_output_params_idx, plan_type
    )
    matches = [
        (th.Tensor, th.Size([100, 784])),
        (th.Tensor, th.Size([100])),
        (th.Tensor, th.Size([10, 100])),
        (th.Tensor, th.Size([10])),
    ]
    assert model_param_type_size == matches

    train_with_hosted_training_plan(
        fl_name, plan_inputs, plan_output_params_idx, plan_type
    )
    accuracy = check_resulting_model(fl_name, model)

    print(f"Model Centric Federated Learning Complete. Accuracy {accuracy:.2F}")
    assert accuracy > 0.05


# === Autograd Plan ===


class MLPAutograd(sy.Module):
    def __init__(self, torch_ref: Any) -> None:
        super().__init__(torch_ref=torch_ref)
        self.l1 = self.torch_ref.nn.Linear(784, 100)
        self.a1 = self.torch_ref.nn.ReLU()
        self.l2 = self.torch_ref.nn.Linear(100, 10)

    def forward(self, x: Any) -> Any:
        l1_out = self.a1(self.l1(x))
        l2_out = self.l2(l1_out)
        return l2_out


def cross_entropy_loss(
    logits: th.Tensor, targets: th.Tensor, batch_size: int
) -> th.Tensor:
    norm_logits = logits - logits.max()
    log_probs = norm_logits - norm_logits.exp().sum(dim=1, keepdim=True).log()
    return -(targets * log_probs).sum() / batch_size


def sgd_step(model: SyModule, lr: float = 0.1) -> None:
    with ROOT_CLIENT.torch.no_grad():
        for p in model.parameters():
            p.data = p.data - lr * p.grad
            p.grad = th.zeros_like(p.grad.get())


def set_params(model: SyModule, params: List) -> None:
    for p, p_new in zip(model.parameters(), params):
        p.data = p_new.data


# === Non-Autograd Plan ===


class MLPNoAutograd(sy.Module):
    """
    Simple model with method for loss and hand-written backprop.
    """

    def __init__(self, torch_ref: Any = th) -> None:
        super(MLPNoAutograd, self).__init__(torch_ref=torch_ref)
        self.fc1 = torch_ref.nn.Linear(784, 100)
        self.relu = torch_ref.nn.ReLU()
        self.fc2 = torch_ref.nn.Linear(100, 10)

    def forward(self, x: Any) -> Any:
        self.z1 = self.fc1(x)
        self.a1 = self.relu(self.z1)
        return self.fc2(self.a1)

    def backward(self, X: Any, error: Any) -> TypeTuple[Any, ...]:
        z1_grad = (error @ self.fc2.state_dict()["weight"]) * (self.a1 > 0).float()
        fc1_weight_grad = z1_grad.t() @ X
        fc1_bias_grad = z1_grad.sum(0)
        fc2_weight_grad = error.t() @ self.a1
        fc2_bias_grad = error.sum(0)
        return fc1_weight_grad, fc1_bias_grad, fc2_weight_grad, fc2_bias_grad

    def softmax_cross_entropy_with_logits(
        self, logits: Any, target: Any, batch_size: int
    ) -> TypeTuple[Any, ...]:
        probs = self.torch_ref.softmax(logits, dim=1)
        loss = -(target * self.torch_ref.log(probs)).mean()
        loss_grad = (probs - target) / batch_size
        return loss, loss_grad

    def accuracy(self, logits: Any, targets: Any, batch_size: int) -> Any:
        pred = self.torch_ref.argmax(logits, dim=1)
        targets_idx = self.torch_ref.argmax(targets, dim=1)
        acc = pred.eq(targets_idx).sum().float() / batch_size
        return acc


def read_file(fname: str) -> str:
    with open(fname, "r") as f:
        return f.read()


private_key = read_file(f"{here}/example_rsa").strip()
public_key = read_file(f"{here}/example_rsa.pub").strip()

auth_token = jwt.encode({}, private_key, algorithm="RS256").decode("ascii")


def create_plan_autograd() -> TypeTuple[Plan, SyModule]:
    local_model = MLPAutograd(th)

    @make_plan
    def train(  # type: ignore
        xs=th.rand([64 * 3, 28 * 28]),
        ys=th.randint(0, 10, [64 * 3, 10]),
        params=List(local_model.parameters()),
    ):

        model = local_model.send(ROOT_CLIENT)
        set_params(model, params)
        for i in range(1):
            indices = th.tensor(range(64 * i, 64 * (i + 1)))
            x, y = xs.index_select(0, indices), ys.index_select(0, indices)
            out = model(x)
            loss = cross_entropy_loss(out, y, 64)
            loss.backward()
            sgd_step(model)

        return (*model.parameters(),)

    return train, local_model


def create_plan_mobile() -> Any:
    def set_remote_model_params(module_ptrs, params_list_ptr):  # type: ignore
        param_idx = 0
        for module_name, module_ptr in module_ptrs.items():
            for param_name, _ in PLAN_BUILDER_VM.store[
                module_ptr.id_at_location
            ].data.named_parameters():
                module_ptr.register_parameter(param_name, params_list_ptr[param_idx])
                param_idx += 1

    local_model = MLPNoAutograd(th)

    # Dummy inputs
    bs = 3
    classes_num = 10
    model_params_zeros = sy.lib.python.List(
        [th.nn.Parameter(th.zeros_like(param)) for param in local_model.parameters()]
    )

    @make_plan
    def training_plan(  # type: ignore
        xs=th.randn(bs, 28 * 28),
        ys=th.nn.functional.one_hot(th.randint(0, classes_num, [bs]), classes_num),
        lr=th.tensor([0.1]),
        batch_size=th.tensor([bs]),
        params=model_params_zeros,
    ):
        # send the model to plan builder (but not its default params)
        model = local_model.send(ROOT_CLIENT, send_parameters=False)

        # set model params from input
        set_remote_model_params(model.modules, params)

        # forward
        logits = model(xs)

        # loss
        loss, loss_grad = model.softmax_cross_entropy_with_logits(
            logits, ys, batch_size
        )

        # backward
        grads = model.backward(xs, loss_grad)

        # SGD step
        updated_params = tuple(
            param - lr * grad for param, grad in zip(model.parameters(), grads)
        )

        # accuracy
        acc = model.accuracy(logits, ys, batch_size)

        # return things
        return (loss, acc, *updated_params)

    # Translate to torchscript
    ts_plan = translate_to_ts(training_plan)

    return training_plan, ts_plan, local_model


def host_to_grid(plans: TypeDict, model: SyModule, name: str) -> None:
    @make_plan
    def avg_plan(  # type: ignore
        avg=List(model.parameters()),
        item=List(model.parameters()),
        num=Int(0),
    ):
        new_avg = []
        for i, param in enumerate(avg):
            new_avg.append((avg[i] * num + item[i]) / (num + 1))
        return new_avg

    name = name

    client_config = {
        "name": name,
        "version": "1.0",
        "batch_size": 64,
        "lr": 0.1,
        "max_updates": 1,  # custom syft.js option that limits number of training loops per worker
    }

    server_config = {
        "min_workers": 2,
        "max_workers": 2,
        "pool_selection": "random",
        "do_not_reuse_workers_until_cycle": 6,
        "cycle_length": 28800,  # max cycle length in seconds
        "num_cycles": 4,  # max number of cycles
        "max_diffs": 1,  # number of diffs to collect before avg
        "minimum_upload_speed": 0,
        "minimum_download_speed": 0,
        "iterative_plan": True,  # tells PyGrid that avg plan is executed per diff
    }

    server_config["authentication"] = {
        "type": "jwt",
        "pub_key": public_key,
    }

    # Auth
    grid_address = f"localhost:{DOMAIN_PORT}"
    grid = ModelCentricFLClient(address=grid_address, secure=False)
    grid.connect()

    # Host
    # If the process already exists, might you need to clear the db.
    grid.host_federated_training(
        model=model,
        client_plans=plans,
        client_protocols={},
        server_averaging_plan=avg_plan,
        client_config=client_config,
        server_config=server_config,
    )


def sanity_check_hosted_plan(
    name: str,
    model: SyModule,
    plan_inputs: OrderedDict,
    plan_output_params_idx: TypeList[int],
    plan_type: str = "list",
) -> TypeList[TypeTuple[type, th.Size]]:
    grid_address = f"localhost:{DOMAIN_PORT}"
    # Authenticate for cycle

    # Helper function to make WS requests
    def sendWsMessage(data: JSONDict) -> JSONDict:
        ws = create_connection("ws://" + grid_address)
        ws.send(json.dumps(data))
        message = ws.recv()
        return json.loads(message)

    auth_request = {
        "type": "model-centric/authenticate",
        "data": {
            "model_name": name,
            "model_version": "1.0",
            "auth_token": auth_token,
        },
    }

    auth_response = sendWsMessage(auth_request)

    # Do cycle request
    cycle_request = {
        "type": "model-centric/cycle-request",
        "data": {
            "worker_id": auth_response["data"]["worker_id"],
            "model": name,
            "version": "1.0",
            "ping": 1,
            "download": 10000,
            "upload": 10000,
        },
    }
    cycle_response = sendWsMessage(cycle_request)
    # Download model

    worker_id = auth_response["data"]["worker_id"]
    request_key = cycle_response["data"]["request_key"]
    model_id = cycle_response["data"]["model_id"]
    training_plan_id = cycle_response["data"]["plans"]["training_plan"]

    def get_model(
        grid_address: str, worker_id: str, request_key: str, model_id: int
    ) -> List:
        req = requests.get(
            (
                f"http://{grid_address}/model-centric/get-model?worker_id={worker_id}&"
                f"request_key={request_key}&model_id={model_id}"
            )
        )
        # TODO migrate to syft-core protobufs
        return deserialize_model_params(req.content)

    # Model
    model_params_downloaded = get_model(grid_address, worker_id, request_key, model_id)

    def get_plan(
        grid_address: str,
        worker_id: int,
        request_key: str,
        plan_id: int,
        plan_type: str,
    ) -> TypeUnion[PlanTorchscript, Plan]:
        req = requests.get(
            (
                f"http://{grid_address}/model-centric/get-plan?worker_id={worker_id}&"
                f"request_key={request_key}&plan_id={plan_id}&receive_operations_as={plan_type}"
            )
        )

        if plan_type == "torchscript":
            pb = PlanTorchscriptPB()
            pb.ParseFromString(req.content)
            return PlanTorchscript._proto2object(pb)
        else:
            pb = PlanPB()
            pb.ParseFromString(req.content)
            return deserialize(pb)

    # Download & Execute Plan
    plan = get_plan(grid_address, worker_id, request_key, training_plan_id, plan_type)
    plan_inputs["params"] = [
        th.nn.Parameter(param) for param in model_params_downloaded
    ]

    if plan_type == "torchscript":
        # kwargs are not supported in torchscript plan
        res = plan(*plan_inputs.values())
    else:
        res = plan(**plan_inputs)

    updated_params = [res[idx] for idx in plan_output_params_idx]

    # Report Model diff
    diff = [orig - new for new, orig in zip(updated_params, model.parameters())]
    diff_serialized = serialize(wrap_model_params(diff)).SerializeToString()

    params = {
        "type": "model-centric/report",
        "data": {
            "worker_id": worker_id,
            "request_key": request_key,
            "diff": base64.b64encode(diff_serialized).decode("ascii"),
        },
    }

    sendWsMessage(params)

    # Check new model
    req_params = {
        "name": name,
        "version": "1.0",
        "checkpoint": "latest",
    }

    res = requests.get(
        f"http://{grid_address}/model-centric/retrieve-model", req_params
    )

    new_model_params = deserialize_model_params(res.content)
    param_type_size = [(type(v), v.shape) for v in new_model_params]

    return param_type_size


def train_with_hosted_training_plan(
    name: str,
    plan_inputs: OrderedDict,
    plan_output_params_idx: TypeList[int],
    plan_type: str = "list",
) -> None:
    # PyGrid Node address
    gridAddress = f"ws://localhost:{DOMAIN_PORT}"

    # Hosted model name/version
    model_name = name
    model_version = "1.0"

    # syft absolute
    # TorchVision hotfix https://github.com/pytorch/vision/issues/3549

    datasets.MNIST.resources = [
        (
            "https://ossci-datasets.s3.amazonaws.com/mnist/train-images-idx3-ubyte.gz",
            "f68b3c2dcbeaaa9fbdd348bbdeb94873",
        ),
        (
            "https://ossci-datasets.s3.amazonaws.com/mnist/train-labels-idx1-ubyte.gz",
            "d53e105ee54ea40749a09fcbcd1e9432",
        ),
        (
            "https://ossci-datasets.s3.amazonaws.com/mnist/t10k-images-idx3-ubyte.gz",
            "9fb629c4189551a2d022fa330f9573f3",
        ),
        (
            "https://ossci-datasets.s3.amazonaws.com/mnist/t10k-labels-idx1-ubyte.gz",
            "ec29112dd5afa0611ce80d1b7f02629c",
        ),
    ]
    datasets.MNIST(get_root_data_path(), train=True, download=True)
    datasets.MNIST(get_root_data_path(), train=False, download=True)

    tfs = transforms.Compose(
        [transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))]
    )
    train_set = datasets.MNIST(
        get_root_data_path(), train=True, download=True, transform=tfs
    )

    cycles_log: TypeList = []
    status = {"ended": False}

    # Called when client is accepted into FL cycle
    def on_accepted(job: FLJob) -> None:
        print(f"Accepted into {job} cycle {len(cycles_log) + 1}.")
        cycle_params = job.client_config
        batch_size, max_updates = (
            cycle_params["batch_size"],
            cycle_params["max_updates"],
        )
        training_plan, model_params = job.plans["training_plan"], job.model
        losses: TypeList = []
        accuracies: TypeList = []

        train_loader = th.utils.data.DataLoader(
            train_set, batch_size=batch_size, drop_last=True, shuffle=True
        )

        for batch_idx, (x, y) in enumerate(train_loader):
            x = x.view(-1, 28 * 28)
            y = th.nn.functional.one_hot(y, 10)
            inputs = plan_inputs
            inputs["xs"] = x
            inputs["ys"] = y
            inputs["params"] = [th.nn.Parameter(param) for param in model_params]
            if plan_type == "torchscript":
                res = training_plan(*inputs.values())
            else:
                res = training_plan(**inputs)
            model_params = [res[idx] for idx in plan_output_params_idx]

            if batch_idx >= max_updates - 1:
                break

        job.report(model_params)

        # Save losses/accuracies from cycle
        cycles_log.append((losses, accuracies))

    # Called when the client is rejected from cycle
    def on_rejected(job: FLJob, timeout: Optional[int] = None) -> None:
        if timeout is None:
            print(f"Rejected from {job} cycle without timeout, FL training complete.")
        else:
            print(f"Rejected from {job} cycle with timeout: {timeout}.")
        status["ended"] = True

    # Called when error occurred
    def on_error(job: FLJob, error: Exception) -> None:
        print(f"Error: {job} {error}")
        status["ended"] = True

    def create_client_and_run_cycle() -> None:
        client = FLClient(url=gridAddress, auth_token=auth_token, secure=False)
        client.worker_id = client.grid_worker.authenticate(
            client.auth_token, model_name, model_version
        )["data"]["worker_id"]
        job = client.new_job(model_name, model_version)

        # Override plan type to use
        job.plan_type = plan_type

        # Set event handlers
        job.add_listener(job.EVENT_ACCEPTED, on_accepted)
        job.add_listener(job.EVENT_REJECTED, on_rejected)
        job.add_listener(job.EVENT_ERROR, on_error)

        # Shoot!
        job.start()

    while not status["ended"]:
        create_client_and_run_cycle()
        time.sleep(1)


def check_resulting_model(name: str, model: SyModule) -> float:
    # Download trained model
    grid_address = f"localhost:{DOMAIN_PORT}"
    grid = ModelCentricFLClient(address=grid_address, secure=False)
    grid.connect()
    trained_params = grid.retrieve_model(name, "1.0")

    # Inference
    def test(test_loader: th.utils.data.DataLoader, model: SyModule) -> th.Tensor:
        correct = []
        model.eval()
        for data, target in test_loader:
            x = data.view(-1, 28 * 28)
            output = model(x)
            _, pred = th.max(output, 1)
            correct.append(th.sum(np.squeeze(pred.eq(target.data.view_as(pred)))))
        acc = sum(correct) / len(test_loader.dataset)
        return acc

    set_params(model, trained_params)

    tfs = transforms.Compose(
        [transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))]
    )
    mnist_test = datasets.MNIST(get_root_data_path(), train=False, transform=tfs)
    test_loader = th.utils.data.DataLoader(
        mnist_test, batch_size=32, shuffle=True, pin_memory=True
    )

    accuracy = test(test_loader, model)

    return accuracy.item()



# stdlib
from dataclasses import dataclass
from enum import Enum
from typing import List


class DriveType(Enum):
    HDD = 0
    SSD = 1


@dataclass
class Drive:
    name: str
    storage: int
    drive_type: DriveType = DriveType.SSD


@dataclass
class Storage:
    drives: List[Drive]

    @property
    def storage_total(self) -> int:
        total = 0
        for d in self.drives:
            total += d.storage
        return total

    @property
    def num_drives(self) -> int:
        return len(self.drives)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/node/common/service/resolve_pointer_type_service.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2

DESCRIPTOR = _descriptor.FileDescriptor(
    name="proto/core/node/common/service/resolve_pointer_type_service.proto",
    package="syft.core.node.common.action",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\nAproto/core/node/common/service/resolve_pointer_type_service.proto\x12\x1csyft.core.node.common.action\x1a%proto/core/common/common_object.proto\x1a\x1bproto/core/io/address.proto"\xc2\x01\n\x19ResolvePointerTypeMessage\x12-\n\x0eid_at_location\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\'\n\x08reply_to\x18\x03 \x01(\x0b\x32\x15.syft.core.io.Address\x12%\n\x06msg_id\x18\x04 \x01(\x0b\x32\x15.syft.core.common.UID"\x83\x01\n\x1fResolvePointerTypeAnswerMessage\x12\x11\n\ttype_path\x18\x01 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12%\n\x06msg_id\x18\x03 \x01(\x0b\x32\x15.syft.core.common.UIDb\x06proto3',
    dependencies=[
        proto_dot_core_dot_common_dot_common__object__pb2.DESCRIPTOR,
        proto_dot_core_dot_io_dot_address__pb2.DESCRIPTOR,
    ],
)


_RESOLVEPOINTERTYPEMESSAGE = _descriptor.Descriptor(
    name="ResolvePointerTypeMessage",
    full_name="syft.core.node.common.action.ResolvePointerTypeMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="id_at_location",
            full_name="syft.core.node.common.action.ResolvePointerTypeMessage.id_at_location",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.core.node.common.action.ResolvePointerTypeMessage.address",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="reply_to",
            full_name="syft.core.node.common.action.ResolvePointerTypeMessage.reply_to",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.core.node.common.action.ResolvePointerTypeMessage.msg_id",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=168,
    serialized_end=362,
)


_RESOLVEPOINTERTYPEANSWERMESSAGE = _descriptor.Descriptor(
    name="ResolvePointerTypeAnswerMessage",
    full_name="syft.core.node.common.action.ResolvePointerTypeAnswerMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="type_path",
            full_name="syft.core.node.common.action.ResolvePointerTypeAnswerMessage.type_path",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="syft.core.node.common.action.ResolvePointerTypeAnswerMessage.address",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="msg_id",
            full_name="syft.core.node.common.action.ResolvePointerTypeAnswerMessage.msg_id",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=365,
    serialized_end=496,
)

_RESOLVEPOINTERTYPEMESSAGE.fields_by_name[
    "id_at_location"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_RESOLVEPOINTERTYPEMESSAGE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_RESOLVEPOINTERTYPEMESSAGE.fields_by_name[
    "reply_to"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_RESOLVEPOINTERTYPEMESSAGE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_RESOLVEPOINTERTYPEANSWERMESSAGE.fields_by_name[
    "address"
].message_type = proto_dot_core_dot_io_dot_address__pb2._ADDRESS
_RESOLVEPOINTERTYPEANSWERMESSAGE.fields_by_name[
    "msg_id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
DESCRIPTOR.message_types_by_name[
    "ResolvePointerTypeMessage"
] = _RESOLVEPOINTERTYPEMESSAGE
DESCRIPTOR.message_types_by_name[
    "ResolvePointerTypeAnswerMessage"
] = _RESOLVEPOINTERTYPEANSWERMESSAGE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ResolvePointerTypeMessage = _reflection.GeneratedProtocolMessageType(
    "ResolvePointerTypeMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _RESOLVEPOINTERTYPEMESSAGE,
        "__module__": "proto.core.node.common.service.resolve_pointer_type_service_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.node.common.action.ResolvePointerTypeMessage)
    },
)
_sym_db.RegisterMessage(ResolvePointerTypeMessage)

ResolvePointerTypeAnswerMessage = _reflection.GeneratedProtocolMessageType(
    "ResolvePointerTypeAnswerMessage",
    (_message.Message,),
    {
        "DESCRIPTOR": _RESOLVEPOINTERTYPEANSWERMESSAGE,
        "__module__": "proto.core.node.common.service.resolve_pointer_type_service_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.node.common.action.ResolvePointerTypeAnswerMessage)
    },
)
_sym_db.RegisterMessage(ResolvePointerTypeAnswerMessage)


# @@protoc_insertion_point(module_scope)



# stdlib
from typing import Any
from typing import Dict

# syft absolute
import syft as sy
from syft.core.io.address import Address
from syft.grid.messages.request_messages import CreateRequestMessage
from syft.grid.messages.request_messages import CreateRequestResponse
from syft.grid.messages.request_messages import DeleteRequestMessage
from syft.grid.messages.request_messages import DeleteRequestResponse
from syft.grid.messages.request_messages import GetRequestMessage
from syft.grid.messages.request_messages import GetRequestResponse
from syft.grid.messages.request_messages import GetRequestsMessage
from syft.grid.messages.request_messages import GetRequestsResponse
from syft.grid.messages.request_messages import UpdateRequestMessage
from syft.grid.messages.request_messages import UpdateRequestResponse


def test_create_data_request_message_serde(node: sy.VirtualMachine) -> None:
    target = Address(name="Alice")

    request_content = {
        "dataset-id": "68a465aer3adf",
        "user-id": "user-id7",
        "request-type": "read",
    }
    msg = CreateRequestMessage(
        address=target,
        content=request_content,
        reply_to=node.address,
    )

    blob = sy.serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_create_data_request_response_serde() -> None:
    target = Address(name="Alice")

    request_content = {"msg": "Request sent succesfully!"}
    msg = CreateRequestResponse(
        address=target,
        status_code=200,
        content=request_content,
    )

    blob = sy.serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_delete_data_request_message_serde(node: sy.VirtualMachine) -> None:
    target = Address(name="Alice")

    request_content = {"request_id": "f2a6as5d16fasd"}
    msg = DeleteRequestMessage(
        address=target,
        content=request_content,
        reply_to=node.address,
    )

    blob = sy.serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_delete_data_request_response_serde() -> None:
    target = Address(name="Alice")

    content = {"msg": "Data Request has been deleted!"}
    msg = DeleteRequestResponse(
        status_code=200,
        address=target,
        content=content,
    )

    blob = sy.serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_update_data_request_message_serde(node: sy.VirtualMachine) -> None:
    target = Address(name="Alice")

    content = {
        "request_id": "546a4d51",
        "dataset-id": "68a465aer3adf",
        "user-id": "user-id7",
        "request-type": "write",
    }
    msg = UpdateRequestMessage(
        address=target,
        content=content,
        reply_to=node.address,
    )

    blob = sy.serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_update_data_request_response_serde() -> None:
    target = Address(name="Alice")

    request_content = {"msg": "Data request has been updated successfully!"}
    msg = UpdateRequestResponse(
        address=target,
        status_code=200,
        content=request_content,
    )

    blob = sy.serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_get_data_request_message_serde(node: sy.VirtualMachine) -> None:
    target = Address(name="Alice")

    content = {"request_id": "eqw9e4a5d846"}
    msg = GetRequestMessage(
        address=target,
        content=content,
        reply_to=node.address,
    )

    blob = sy.serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_get_data_request_response_serde() -> None:
    target = Address(name="Alice")

    content = {
        "request_id": "asfdaead131",
        "dataset-id": "68a465aer3adf",
        "user-id": "user-id7",
        "request-type": "read",
    }

    msg = GetRequestResponse(
        address=target,
        status_code=200,
        content=content,
    )

    blob = sy.serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_get_all_data_requests_message_serde(node: sy.VirtualMachine) -> None:
    target = Address(name="Alice")

    content: Dict[Any, Any] = {}
    msg = GetRequestsMessage(
        address=target,
        content=content,
        reply_to=node.address,
    )

    blob = sy.serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2


def test_get_all_data_requests_response_serde() -> None:
    target = Address(name="Alice")

    request_content = {
        "workers": {
            "626sadaf631": {
                "dataset-id": "68a465aer3adf",
                "user-id": "user-id7",
                "request-type": "read",
            },
            "a84ew64wq6e": {
                "dataset-id": "98w4e54a6d",
                "user-id": "user-id9",
                "request-type": "write",
            },
        }
    }

    msg = GetRequestsResponse(
        address=target,
        status_code=200,
        content=request_content,
    )

    blob = sy.serialize(msg)
    msg2 = sy.deserialize(blob=blob)

    assert msg.id == msg2.id
    assert msg.address == target
    assert msg.content == msg2.content
    assert msg == msg2



# third party
import torch as th

# syft absolute
import syft as sy
from syft.core.common.uid import UID
from syft.core.node.common.action.function_or_constructor_action import (
    RunFunctionOrConstructorAction,
)


# TODO test execution
# TODO test permissions
def test_run_function_or_constructor_action_serde(
    root_client: sy.VirtualMachineClient,
) -> None:
    args = (
        th.tensor([1, 2, 3]).send(root_client),
        th.tensor([4, 5, 5]).send(root_client),
    )

    msg = RunFunctionOrConstructorAction(
        path="torch.Tensor.add",
        args=args,
        kwargs={},
        id_at_location=UID(),
        address=root_client.address,
        msg_id=UID(),
    )

    blob = sy.serialize(msg)

    msg2 = sy.deserialize(blob=blob)

    assert msg2.path == msg.path
    # FIXME this cannot be checked before we fix the Pointer serde problem (see _proto2object in Pointer)
    # assert msg2.args == msg.args
    assert msg2.kwargs == msg.kwargs
    assert msg2.address == msg.address
    assert msg2.id == msg.id
    assert msg2.id_at_location == msg.id_at_location



# stdlib
import binascii
import json
from typing import Any as TypeAny
from typing import Dict as TypeDict
from typing import Optional
from typing import Union as TypeUnion

# third party
from google.protobuf.message import Message
from google.protobuf.reflection import GeneratedProtocolMessageType
import requests
import websocket

# syft relative
from ..core.common.serde.deserialize import _deserialize as deserialize
from ..core.common.serde.serialize import _serialize as serialize
from ..federated import JSONDict

TIMEOUT_INTERVAL = 60


class GridError(BaseException):
    def __init__(
        self, error: TypeUnion[Exception, str], status: Optional[int] = None
    ) -> None:
        super().__init__(error)
        if type(error) is not str:
            error = str(error)
        self.error = error
        self.status = status


class ModelCentricFLBase:
    def __init__(self, address: str, secure: bool = False):
        self.address = address
        self.secure = secure
        self.ws: Optional[websocket._core.WebSocket] = None

    @property
    def ws_url(self) -> str:
        protocol = "wss" if self.secure else "ws"
        return f"{protocol}://{self.address}"

    @property
    def http_url(self) -> str:
        protocol = "https" if self.secure else "http"
        return f"{protocol}://{self.address}"

    def connect(self) -> None:
        args_ = {"max_size": None, "timeout": TIMEOUT_INTERVAL, "url": self.ws_url}
        self.ws = websocket.create_connection(**args_)

    def _send_msg(self, message: TypeDict[str, TypeAny]) -> JSONDict:
        """Prepare/send a JSON message to a PyGrid server and receive the response.

        Args:
            message (dict) : message payload.
        Returns:
            response (dict) : response payload.
        """
        if self.ws is None or not self.ws.connected:
            self.connect()

        if self.ws is not None:
            self.ws.send(json.dumps(message))
            json_response = json.loads(self.ws.recv())

            # Look for error in root and under "data"
            error = None
            if "data" in json_response:
                error = json_response["data"].get("error", None)
            elif "error" in json_response:
                error = json_response["error"]

            if error is not None:
                raise GridError(error, None)

            return json_response
        else:
            raise GridError("Websocket connection unavailable", None)

    def _send_http_req(
        self,
        method: str,
        path: str,
        params: Optional[JSONDict] = None,
        body: Optional[JSONDict] = None,
    ) -> bytes:
        if method == "GET":
            res = requests.get(self.http_url + path, params)
        elif method == "POST":
            res = requests.post(self.http_url + path, params=params, data=body)

        if not res.ok:
            error = "HTTP response is not OK"
            try:
                json_response = json.loads(res.content)
                error = json_response.get("error", error)
            finally:
                raise GridError(f"Grid Error: {error}", res.status_code)

        return res.content

    def _serialize(self, obj: object) -> bytes:
        """Serializes object to protobuf"""
        pb: Message = serialize(obj)  # type: ignore
        return pb.SerializeToString()

    def _serialize_dict_values(self, obj: JSONDict) -> JSONDict:
        serialized_object = {}
        for k, v in obj.items():
            serialized_object[k] = binascii.hexlify(self._serialize(v)).decode()
        return serialized_object

    def _unserialize(
        self, serialized_obj: bytes, obj_protobuf_type: GeneratedProtocolMessageType
    ) -> TypeAny:
        pb = obj_protobuf_type()
        pb.ParseFromString(serialized_obj)
        return deserialize(pb)

    def close(self) -> None:
        if self.ws is not None:
            self.ws.shutdown()

    def hex_serialize(self, x: object) -> str:
        return binascii.hexlify(self._serialize(x)).decode()



# stdlib
from dataclasses import asdict

# syft absolute
from syft.core.node.device.device_type.device_type import DeviceType
from syft.core.node.device.device_type.specs.cpu import CPU
from syft.core.node.device.device_type.specs.cpu import CpuArchitectureTypes
from syft.core.node.device.device_type.specs.gpu import GPU
from syft.core.node.device.device_type.specs.network import Network
from syft.core.node.device.device_type.specs.provider import Provider
from syft.core.node.device.device_type.specs.storage import Drive
from syft.core.node.device.device_type.specs.storage import DriveType
from syft.core.node.device.device_type.specs.storage import Storage


def test_gpu_device_type() -> None:

    mbs = 16 * 1024 * 1024
    gpu = GPU(
        name="V100",
        count=1,
        manufacturer="NVIDIA",
        memory_total=mbs,
        memory_per_gpu=mbs,
    )

    fixture = {
        "name": "V100",
        "count": 1,
        "manufacturer": "NVIDIA",
        "memory_total": 16777216,
        "memory_per_gpu": 16777216,
    }

    assert fixture == asdict(gpu)


def test_cpu_device_type() -> None:

    cpu = CPU(
        num_cores=8,
        num_threads_per_core=2,
        clock_speed=2.3,
        architecture=CpuArchitectureTypes.x86_64,
    )

    fixture = {
        "num_cores": 8,
        "num_threads_per_core": 2,
        "clock_speed": 2.3,
        "architecture": CpuArchitectureTypes.x86_64,
    }

    assert fixture == asdict(cpu)


def test_network_device_type() -> None:
    net = Network(performance="1000MB")
    fixture = {"performance": "1000MB", "ipv6_supported": False}

    assert fixture == asdict(net)


def test_storage_drives() -> None:
    mbs = 1 * 1024 * 1024 * 1024
    drive = Drive(name="Macintosh HD", storage=mbs)
    raid = Storage(drives=[drive, drive])

    drive_fixture = {
        "name": "Macintosh HD",
        "storage": 1073741824,
        "drive_type": DriveType.SSD,
    }
    raid_fixture = {"drives": [drive_fixture, drive_fixture]}

    assert drive_fixture == asdict(drive)
    assert raid_fixture == asdict(raid)
    assert raid.storage_total == raid.num_drives * mbs
    assert raid.num_drives == 2


def test_provider() -> None:
    print("gfds", str(Provider.USER_OWNED))
    assert Provider.USER_OWNED.value == "user-owned"
    assert Provider.AWS.value == "aws"
    assert Provider.AZURE.value == "azure"
    assert Provider.GCP.value == "gcp"
    assert Provider.GENESIS.value == "genesis"


def test_device_type() -> None:
    cpu = CPU(
        num_cores=8,
        num_threads_per_core=2,
        clock_speed=2.3,
        architecture=CpuArchitectureTypes.x86_64,
    )

    mbs = 1 * 1024 * 1024 * 1024
    drive = Drive(name="Macintosh HD", storage=mbs)
    raid = Storage(drives=[drive, drive])

    macbook_pro_16in_2019 = DeviceType(
        name="MacBook Pro (16-inch, 2019)",
        provider=Provider.USER_OWNED,
        price_per_hour=0,
        memory=32768,
        bare_metal=True,
        storage=raid,
        cpu=cpu,
        gpu=None,
        network=None,
        spot_mode_supported=False,
        price_per_hour_spot=0,
    )

    laptop_fixture = {
        "name": "MacBook Pro (16-inch, 2019)",
        "provider": Provider.USER_OWNED,
        "price_per_hour": 0,
        "memory": 32768,
        "bare_metal": True,
        "storage": {
            "drives": [
                {
                    "name": "Macintosh HD",
                    "storage": 1073741824,
                    "drive_type": DriveType.SSD,
                },
                {
                    "name": "Macintosh HD",
                    "storage": 1073741824,
                    "drive_type": DriveType.SSD,
                },
            ]
        },
        "cpu": {
            "num_cores": 8,
            "num_threads_per_core": 2,
            "clock_speed": 2.3,
            "architecture": CpuArchitectureTypes.x86_64,
        },
        "gpu": None,
        "network": None,
        "spot_mode_supported": False,
        "price_per_hour_spot": 0,
    }

    assert laptop_fixture == asdict(macbook_pro_16in_2019)



# stdlib
from typing import Any

# third party
import pytest
import torch as th

# syft absolute
import syft as sy
from syft.core.common.uid import UID
from syft.lib.python import String


def test_device() -> None:
    device = th.device("cuda")
    assert device.type == "cuda"
    assert device.index is None


def test_device_init(node: sy.VirtualMachine, client: sy.VirtualMachineClient) -> None:
    assert node.name == "Bob"
    torch = client.torch

    type_str = String("cuda:0")
    str_pointer = type_str.send(client)

    device_pointer = torch.device(str_pointer)
    assert type(device_pointer).__name__ == "devicePointer"
    assert isinstance(device_pointer.id_at_location, UID)


@pytest.mark.slow
@pytest.mark.parametrize("type_str", ["cpu", "cuda"])
@pytest.mark.parametrize("index", [None, 0])
def test_device_serde(
    type_str: str, index: Any, root_client: sy.VirtualMachineClient
) -> None:
    device = th.device(type_str, index)
    device_ptr = device.send(root_client)
    assert device_ptr.get() == device



# stdlib
from typing import Any

# syft absolute
import syft as sy

alice = sy.VirtualMachine(name="alice")
alice_client = alice.get_root_client()
remote_python = alice_client.syft.lib.python


def get_permission(obj: Any) -> None:
    remote_obj = alice.store[obj.id_at_location]
    remote_obj.read_permissions[alice_client.verify_key] = obj.id_at_location


x_data = list(range(13))
y_data = [1, 3, 2, 5, 7, 8, 8, 9, 10, 12, 12, 13, 15]
zip(x_data, y_data)
x_data_ptr = sy.lib.python.List(x_data).send(alice_client)
y_data_ptr = sy.lib.python.List(y_data).send(alice_client)
print("gata")
# n_ptr = x_data_ptr.__len__()
#
# x_mean_ptr = sum(x_data_ptr) / n_ptr
# y_mean_ptr = sum(y_data_ptr) / n_ptr
#
# crossdev_list_ptr = remote_python.List()
# dev_list_ptr = remote_python.List()
#
# for x_ptr, y_ptr in zip(x_data_ptr, y_data_ptr):
#     crossdev_list_ptr.append(x_ptr * y_ptr)
#     dev_list_ptr.append(x_ptr * x_ptr)
#
#
# cross_dev_ptr = sum(crossdev_list_ptr) - n_ptr * x_mean_ptr * y_mean_ptr
# dev_ptr = sum(dev_list_ptr) - n_ptr * x_mean_ptr * x_mean_ptr
#
# b_1_ptr = cross_dev_ptr / dev_ptr
# b_0_ptr = y_mean_ptr - b_1_ptr * x_mean_ptr
#
# get_permission(b_1_ptr)
# get_permission(b_0_ptr)
#
# b_1 = b_1_ptr.get()
# b_0 = b_0_ptr.get()
#
# print(b_1)
# print(b_0)
